#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# @Time    : 2024/3/7
# @Author  : yanxiaodong
# @File    : metric.py
"""
import os
from typing import List, Union, Dict, Optional
from collections import defaultdict

from gaea_tracker import ExperimentTracker
from pydantic import BaseModel
from windmillclient.client.windmill_client import WindmillClient
from windmilltrainingv1.client.training_api_project import parse_project_name

from gaea_metric.image import MeanAveragePrecision, Accuracy, PrecisionRecallF1score, ConfusionMatrix, MeanIoU
from gaea_operator.utils import read_file, format_time, write_file, DEFAULT_METRIC_FILE_NAME


class Label(BaseModel):
    """
    Labeled Object
    """
    id: int
    name: str


class BaseMetric(BaseModel):
    """
    Metric Object
    """
    modelName: Optional[str] = None
    datasetName: Optional[str] = None
    baselineJobName: Optional[str] = None
    createdAt: Optional[str] = None
    updatedAt: Optional[str] = None


class Metric(object):
    """
    Metric class
    """

    def __init__(self,
                 category: str,
                 labels: List[Label],
                 images: List[Dict],
                 windmill_client: WindmillClient = None,
                 tracker_client: ExperimentTracker = None):
        self.windmill_client = windmill_client
        self.tracker_client = tracker_client

        self.labels = labels
        self.images = images
        self.image_dict = {item["image_id"]: item for item in self.images}

        if category == "Image/ObjectDetection":
            self.metric_name = [MeanAveragePrecision(categories=self.labels,
                                                     num_classes=len(self.labels),
                                                     classwise=True)]
            self.format_function = self._format_to_object_detection
        elif category == "Image/Classification":
            self.metric_name = [Accuracy(num_classes=len(self.labels)),
                                PrecisionRecallF1score(num_classes=len(self.labels)),
                                ConfusionMatrix(num_classes=len(self.labels))]
            self.format_function = None
        elif category == "Image/SemanticSegmentation":
            self.metric_name = [MeanIoU(num_classes=len(self.labels))]
            self.format_function = None
        else:
            raise ValueError(f"Unknown category: {category}")
        self.category = category

    def update(self, predictions: List[Dict], references: List[Dict]):
        """
        Update metric.
        """
        predictions = self.format_function(predictions=predictions)
        references = self.format_function(references=references)

        for metric in self.metric_name:
            metric.update(predictions=predictions, references=references)

    def _format_to_object_detection(self, raw_data: List[Dict]):
        """
        Format to object detection metric.
        """
        annos = defaultdict(list)
        for item in raw_data:
            im_id = item["image_id"]
            img = self.image_dict[im_id]
            for anno in item["annotations"]:
                anno["width"] = img["width"]
                anno["height"] = img["height"]
                anno['ignore'] = anno['ignore'] if 'ignore' in anno else 0
                anno['ignore'] = "iscrowd" in anno and anno["iscrowd"]
                anno['category_id'] = anno["labels"][0]["id"]
                anno['score'] = anno["labels"][0]["score"]
                annos[im_id].append(anno)

    def _format_object_detection_result(self, metric_result: Dict):
        return metric_result

    def _format_to_classification(self, predictions: List[Dict], references: List[Dict]):
        """
        Format to classification metric.
        """
        pass

    def compute(self):
        """
        Compute metric.
        """
        results = {}
        for metric in self.metric_name:
            results[metric.name] = metric.compute()

        return results

    def save(self, metric_result: Dict, output_uri: str = None):
        """
        Save metric.
        """
        if output_uri is None:
            output_uri = os.path.join(self.tracker_client.job_work_dir, DEFAULT_METRIC_FILE_NAME)

        metric_result = self._format_object_detection_result(metric_result=metric_result)

        write_file(obj=metric_result, output_dir=os.path.dirname(output_uri), file_name=os.path.basename(output_uri))

    def update_metric_file(self, output_dir: str, dataset_name: str, model_object_name: str, model_artifact_name: str):
        """
        Update metric file.
        """
        response = self.windmill_client.get_artifact(object_name=model_object_name, version="best")

        baseline_model_name = model_artifact_name
        if getattr(response, "name"):
            if response.name != model_artifact_name:
                baseline_model_name = response.name
        base_metric = BaseMetric(modelName=model_artifact_name, datasetName=dataset_name, updatedAt=format_time())

        tags = [{"modelName": baseline_model_name}, {"datasetName": dataset_name}]
        workspace_id, project_name = parse_project_name(self.tracker_client.project_name)
        response = self.windmill_client.list_job(workspace_id=workspace_id, project_name=project_name, tags=tags)
        if len(response.result) > 0:
            base_metric.baselineJobName = response.result[0]["name"]
        metric_data = read_file(input_dir=output_dir, file_name=DEFAULT_METRIC_FILE_NAME)
        metric_data.update(base_metric.dict())
        write_file(obj=metric_data, output_dir=output_dir, file_name=DEFAULT_METRIC_FILE_NAME)

    def __call__(self, predictions: List[Dict], references: List[Dict], output_uri: str):
        self.update(predictions=predictions, references=references)
        results = self.compute()
        self.save(metric_result=results, output_uri=output_uri)


def get_score_from_file(filepath: str, metric_name: str):
    """
    Get metric name score from file.
    """
    metric_data = read_file(input_dir=os.path.dirname(filepath), file_name=os.path.basename(filepath))
    for metric in metric_data["metrics"]:
        if metric["name"] == metric_name:
            if isinstance(metric["result"], Dict):
                return list(metric["result"].values())[0]
            return metric["result"]